#!/usr/bin/env python
#@+leo-ver=4
#@+node:@file flickrfs.py
# v0.9 - Initial release
# v1.0 - Features added
# v1.1 -
#      - fixed bug if photo title=='', then flickrfs crashes. 
#      - Including API Key
#      - Support for spaces in tags while uploading
#      - Permissions support while Uploading (mode=abc - a=7(always),
#				b=(friends=5 | family=6 | none=4), c=(5=public | 4=private)
#===============================================================================
#    Copyright (C) 2005 Manish Rai Jain  <manishrjain@gmail.com>
#
#    This program can be distributed under the terms of the GNU GPL version 2, or 
#    its later versions. 
#
# DISCLAIMER: The API Key and Shared Secret are provided by the author in 
# the hope that it will prevent unnecessary trouble to the end-user. The 
# author will not be liable for any misuse of this API Key/Shared Secret 
# through this application/derived apps/any 3rd party apps using this key. 
#===============================================================================
#
#@+others
#@+node:imports

from fuse import Fuse
import os
from errno import *
from stat import *

import thread, array, string, urllib2, traceback
#@-node:imports
#@+node:class Xmp

#Some global definitions and functions

# Setup logging
import time, logging, logging.handlers
log = logging.getLogger('flickrfs')

loghdlr = logging.handlers.RotatingFileHandler("/var/log/flickrfs", "a", 5242880, 3)
logfmt = logging.Formatter("%(asctime)s %(levelname)-10s %(message)s", "%x %X")
loghdlr.setFormatter(logfmt)
log.addHandler(loghdlr)
log.setLevel(logging.DEBUG)

#Import flickr python api
from flickrapi import FlickrAPI

# flickr auth information
flickrAPIKey = "f8aa9917a9ae5e44a87cae657924f42d"  # API key
flickrSecret = "3fbf7144be7eca28"                  # shared "secret"
browserName = "/usr/bin/firefox"                   # for out-of-band auth inside a web browser


class TransFlickr:  #Transactions with flickr
	def uploadfile(self, filepath, taglist, bufData, mode):
		public = mode&1 #Set public 4(always), 1(public). Public overwrites f&f
		friends = mode>>3 & 1 #Set friends and family 4(always), 2(family), 1(friends)
		family = mode>>4 & 1
		log.info("Uploading file: " + filepath + ":with data of len:" + str(len(bufData)))
		log.info("Permissions:Family:%s Friends:%s Public:%s"%(family,friends,public))
		rsp = fapi.upload(filename=filepath, jpegData=bufData, api_key=flickrAPIKey, auth_token=token, \
        	        title=os.path.splitext(os.path.basename(filepath))[0], \
                	tags=taglist, \
                	is_public=public and "1" or "0", \
                	is_friend=friends and "1" or "0", \
                	is_family=family and "1" or "0")

		if rsp==None:
			log.error("Can't write file: " + filepath)
		else:
			retinfo = fapi.returntestFailure(rsp)
			if retinfo=="OK":
				id = rsp.photoid[0].elementText
				log.info("File uploaded:" + filepath + ":with photoid:" + id + ":")
				return id
			else:
				log.error(retinfo)
	def put2Set(self, set_id, photo_id):
		log.info("Uploading photo:"+photo_id+":to set_id:"+set_id)
		rsp = fapi.photosets_addPhoto(api_key=flickrAPIKey, auth_token=token,\
				photoset_id=set_id, photo_id=photo_id)

		retinfo = fapi.returntestFailure(rsp)
		if retinfo=="OK":
			log.info("Uploaded photo to set")
		else:
			log.error(retinfo)
	
	def createSet(self, path, photo_id):
		log.info("Creating set:%s:with primary photo:%s:"%(path,photo_id))
		ind = path.rindex('/')
		title = path[ind+1:]
		rsp = fapi.photosets_create(api_key=flickrAPIKey, auth_token=token,\
			title=title, primary_photo_id=photo_id)
		retinfo = fapi.returntestFailure(rsp)
		if retinfo=="OK":
			log.info("Created set:%s:"%(title))
			return rsp.photoset[0]['id']
		else:
			log.error(retinfo)
	
	def deleteSet(self, set_id):
		log.info("Deleting set:%s:"%(set_id))
		if str(set_id)=="0":
			log.info("The set is not existant online.")
			return
			
		rsp = fapi.photosets_delete(api_key=flickrAPIKey, auth_token=token,\
			photoset_id=set_id)
		retinfo = fapi.returntestFailure(rsp)
		if retinfo=="OK":
			log.info("Deleted set")
		else:
			log.error(retinfo)
	
	def getPhotoInfo(self, photoId):
		rsp = fapi.photos_getInfo(photo_id=photoId, api_key=flickrAPIKey, auth_token=token)
		retinfo = fapi.returntestFailure(rsp)
		if retinfo!="OK":
			log.error("Can't retrieve information about photo: " + photoId)
			return None
		return (0,0,0,1,0,0,0,0,0,0)

	def getPhoto(self, photoId):
		rsp = fapi.photos_getSizes(photo_id=photoId, api_key=flickrAPIKey, auth_token=token)
		retinfo = fapi.returntestFailure(rsp)
		if retinfo!="OK":
			log.error("Can't get information about photo: " + photoId)
			return None
		buf = ""
		for a in rsp.sizes[0].size:
			if a['label']=='Large':
				try:
					f = urllib2.urlopen(a['source'])
					buf = f.read()
				except KeyError:
					return ""
				except:
					return ""
		return buf
	def removePhotofromSet(self, photoId, photosetId):
		rsp = fapi.photosets_removePhoto(api_key=flickrAPIKey, auth_token=token,\
			photo_id=photoId, photoset_id=photosetId)
		retinfo = fapi.returntestFailure(rsp)
		if retinfo=="OK":
			log.info("Photo removed from set")
		else:
			log.error(retinfo)
			
		
	def getBandwidthInfo(self):
		log.debug("Retrieving bandwidth information")
		rsp = fapi.people_getUploadStatus(api_key=flickrAPIKey, auth_token=token)
		retinfo = fapi.returntestFailure(rsp)
		bw = rsp.user[0].bandwidth[0]
		log.debug("Bandwidth: max:" + bw['max'])
		log.debug("Bandwidth: used:" + bw['used'])
		if retinfo=="OK":
			return (bw['max'], bw['used'])
		else:
			log.error("Can't retrieve bandwidth information")
			return (None,None)
		
class FlickrInode:
	"""Class used to store Flickrfs inodes
	"""
	def __init__(self, path = None, ID = "", isDir=False, MODE = 0):
		if isDir:
			if MODE==0:
				self.mode = S_IFDIR | 0755
			else:
				self.mode = S_IFDIR | MODE
			self.nlink = 2
			self.size = 1L
			self.setId = ID
			self.dirfile = ""
		else:
			if MODE==0:
				self.mode = S_IFREG | 0644
			else:
				self.mode = S_IFREG | MODE
			self.nlink = 1
			self.size = 1L
			self.buf = ""
			self.photoId = ID

		timeInst = int(time.time())
		self.ino = long(time.time())
		self.dev = 409089L
		self.uid = int(os.getuid())
		self.gid = int(os.getgid())
		self.atime = timeInst
		self.mtime = timeInst
		self.ctime = timeInst
		self.blocksize = DefaultBlockSize

		

class Flickrfs(Fuse):

	#@@+others
    
	#@+node:__init__
    	def __init__(self, *args, **kw):
    
        	Fuse.__init__(self, *args, **kw)
    
        	if 1:
            		log.info("flickrfs.py:Flickrfs:mountpoint: %s" % repr(self.mountpoint))
            		log.info("flickrfs.py:Flickrfs:unnamed mount options: %s" % self.optlist)
	       		log.info("flickrfs.py:Flickrfs:named mount options: %s" % self.optdict)
	    		log.info("Authorizing with flickr...")
	    		# initialise FlickrAPI object
		
		self.inodeCache = {}  #Cached inodes for faster access
#		self.listDir = {} #Store the file names inside the directory
		self.NSID = ""
		global fapi
    		global token
		global DefaultBlockSize
		DefaultBlockSize = 4*1024  #4KB
		fapi = FlickrAPI(flickrAPIKey, flickrSecret)

		# proceed with auth
		# TODO use auth.checkToken function if available, and wait after opening browser
		token = fapi.getToken(browser = browserName)
		if token == None:
			log.error('Not able to authorize. Exiting...')
			sys.exit(-1)
	    		#Add some authorization checks here(?)
		log.info('Authorization complete')
        	# do stuff to set up your filesystem here, if you want
		self._mkfileOrDir("/", isDir=True)
		self._mkfileOrDir("/tags", isDir=True)
		self._mkfileOrDir("/tags/personal", isDir=True)
		self._mkfileOrDir("/tags/public", isDir=True)
        	thread.start_new_thread(self.sets_thread, ())

	#@+node:sets_thread
	def sets_thread(self):
		"""
        	The beauty of the FUSE python implementation is that with the python interp
        	running in foreground, you can have threads
        	"""    
		log.info("sets_thread: started")
		self._mkfileOrDir("/sets", isDir=True)
		rsp = fapi.photosets_getList(api_key=flickrAPIKey, auth_token=token)
		if rsp!=None:
			for a in rsp.photosets[0].photoset:
				title = a.title[0].elementText.replace('/', ' ')
				curdir = "/sets/" + title
				if title.strip()=='':
					curdir = "/sets/" + a['id']
				set_id = a['id']
				self._mkfileOrDir(curdir, id=set_id, isDir=True)
				photos = fapi.photosets_getPhotos(api_key=flickrAPIKey, photoset_id=set_id)
				if photos!=None:
					for b in photos.photoset[0].photo:
						title = b['title'].replace('/', ' ')
						if title.strip()=='':
							title = str(b['id'])
						title = title[:32]   #Only allow 32 characters
						self._mkfileOrDir(curdir+'/'+title, \
							id=str(b['id']), isDir=False)
					
					
				

	#@-node:sets_thread
	
	def tags_thread(self, path):
		ind = string.rindex(path, '/')
		tagName = path[ind+1:]
		if tagName.strip()=='':
			log.error("The tagName:%s: doesn't contain any tags"%(tagName))
			return 

		log.info("tags_thread:" + tagName + ":started")
		sendtagList = ','.join(tagName.split(':'))

		if(path.startswith('/tags/personal')):
			if self.NSID=="":
				tags_rsp = fapi.auth_checkToken(api_key=flickrAPIKey, auth_token=token)
				if tags_rsp!=None:
					usr = tags_rsp.auth[0].user[0]
					self.NSID = usr['nsid']
					log.debug("Got NSID:"+ self.NSID+ ":")
			tags_rsp = fapi.photos_search(api_key=flickrAPIKey,user_id=self.NSID,\
				tags=sendtagList, tag_mode="all")
		elif(path.startswith('/tags/public')):
			tags_rsp = fapi.photos_search(api_key=flickrAPIKey, tags=sendtagList,\
				tag_mode="all")
		else:
			return

		log.debug("Search for photos with tag:" + sendtagList + ":done")
		fapi.testFailure(tags_rsp)
		if tags_rsp!=None:
			for a in tags_rsp.photos:
#				curdir = "/tags/" + str(a['page']) #Can't do this. Will create recursion
#				self._mkfileOrDir(curdir)
#				log.debug(str(a['page']))
				for b in a.photo:
					try:
						title = b['title'].replace('/', ' ')
						if title.strip()=='':
							title = str(b['id'])
						title = title[:32]   #Only allow 32 characters
						self._mkfileOrDir(path +"/" + title,\
								id=str(b['id']), isDir=False)
						counter += 1
					except:
						pass
	
	#@+node:attribs
    	flags = 1
	#@-node:attribs

	def _mkfileOrDir(self, pth, id="", isDir=False, MODE=0):
		path = str(pth)
		ind = string.rindex(path, '/')
		parentDir = path[:ind]
		if parentDir=='':
			parentDir = '/'
		log.debug("parentDir:" + parentDir + ":")
		if isDir==True:
			log.debug("Creating directory:" + path)
			self.inodeCache[path] = FlickrInode(path, ID=id, isDir=True)
			if path=='/':
				log.debug("This is root already. Can't find parent of GOD!!!")
			else:
				pinode = self.getInode(parentDir)
				pinode.nlink += 1
		else:
			log.debug("Creating file:" + path + ":with id:" + id)
			self.inodeCache[path] = FlickrInode(path, ID=id, isDir=False, MODE=MODE)
	
	#@+node:getattr
    	def getattr(self, path):
    		log.debug("getattr:" + path + ":")
		if path.startswith('/sets/'):
			templist = path.split('/')
			ind = templist.index('sets')
			setName = templist[ind+1].split(':')[0]
			templist[2] = setName
			path = '/'.join(templist)
			log.debug("getattr:After modifying:%s:" % (path))

		inode=self.getInode(path)
        	if inode:
			log.debug("inode "+str(inode))
            		statTuple = (inode.mode,inode.ino,inode.dev,inode.nlink, \
				inode.uid,inode.gid,inode.size,inode.atime,inode.mtime,inode.ctime)
                	log.debug("statsTuple "+str(statTuple))
            		return statTuple
        	else:
			e = OSError("No such file"+path)
            		e.errno = ENOENT
           		raise e

	#@-node:getattr

	#@+node:readlink
    	def readlink(self, path):
    		log.debug("readlink")
    		return os.readlink(path)
    	#@-node:readlink
	
    	#@+node:getdir
    	def getdir(self, path):
    		log.debug("getdir:" + path)
#		return map(lambda x: (x,0),  self.listDir[path])
		templist = ['.', '..']
		for a in self.inodeCache.keys():
			ind = a.rindex('/')
			if path=='/':
				path=""
			if path==a[:ind]:
				name = a.split('/')[-1]
				if name!="":
					templist.append(name)
		return map(lambda x: (x,0), templist)

    	#@-node:getdir
	
    	#@+node:unlink
    	def unlink(self, path):
    		log.debug("unlink:%s:" % (path))
		if self.inodeCache.has_key(path):
			inode = self.inodeCache.pop(path)
			if path.startswith('/sets/'):
				ind = path.rindex('/')
				pPath = path[:ind]
				pinode = self.getInode(pPath)
				TransFlickr().removePhotofromSet(photoId=inode.photoId, photosetId=pinode.setId)
	
			del inode
		else:
			log.error("Can't find what you want to remove")
			e = OSError("No such file"+path)
            		e.errno = ENOENT
           		raise e
    	#@-node:unlink
	
    	#@+node:rmdir
    	def rmdir(self, path):
		log.debug("rmdir:%s:"%(path))
		if self.inodeCache.has_key(path):
			for a in self.inodeCache.keys():
				if a.startswith(path+'/'):
					e = OSError("Directory not empty")
					e.errno = ENOTEMPTY
					raise e
		else:
			log.error("Can't find the directory you want to remove")
			e = OSError("No such folder"+path)
            		e.errno = ENOENT
           		raise e
			
		ind = path.rindex('/')
		pPath = path[:ind]
		if path.startswith('/tags/personal/'):	
			inode = self.inodeCache.pop(path)
			del inode
			pInode = self.getInode(pPath)
			pInode.nlink -= 1
		elif path.startswith('/tags/public/'):
			inode = self.inodeCache.pop(path)
			del inode
			pInode = self.getInode(pPath)
			pInode.nlink -= 1
		elif path.startswith('/sets/'):
			inode = self.inodeCache.pop(path)
			TransFlickr().deleteSet(inode.setId)
			del inode
			pInode = self.getInode(pPath)
			pInode.nlink -= 1
		else:
    			log.debug("rmdir!! I refuse to do anything! <Stubborn>")
			e = OSError("Removal of folder %s not allowed" % (path))
			e.errno = EPERM
			raise e
			
	#    	return os.rmdir(path)	
	#@-node:rmdir
	
    	#@+node:symlink
    	def symlink(self, path, path1):
    		log.debug("symlink")
    		return os.symlink(path, path1)	
    	#@-node:symlink
    
    	#@+node:rename
    	def rename(self, path, path1):
    		log.debug("rename?? My parents named me. Not changing!!")
#    		return os.rename(path, path1)
    	#@-node:rename
    
    	#@+node:link
    	def link(self, path, path1):
    		log.debug("link")
#    		return os.link(path, path1)
    	#@-node:link
    
    	#@+node:chmod
    	def chmod(self, path, mode):
    		log.debug("chmod. Haha! No use!")
#    		return os.chmod(path, mode)
    	#@-node:chmod
    
    	#@+node:chown
    	def chown(self, path, user, group):
    		log.debug("chown. will support later")
#    		return os.chown(path, user, group)
    	#@-node:chown
    
    	#@+node:truncate
    	def truncate(self, path, size):
    		log.debug("truncate?? WTF for?")
#    		f = open(path, "w+")
#    		return f.truncate(size)
    	#@-node:truncate
    
    	#@+node:mknod
    	def mknod(self, path, mode, dev):
    		""" Python has no os.mknod, so we can only do some things """
		log.debug("mknod? OK! Had a close encounter!!")
		if path.startswith('/sets/'):
			templist = path.split('/')
			ind = templist.index('sets')
			setName = templist[ind+1].split(':')[0]
			templist[2] = setName
			path = '/'.join(templist)
			log.debug("mknod:After modifying:%s:" % (path))

		self._mkfileOrDir(path, id="NEW", isDir=False, MODE=mode)
#    		if S_ISREG(mode):
#    			open(path, "w")
#    		else:
#    			return -EINVAL
    	#@-node:mknod
    	#@+node:mkdir
    	def mkdir(self, path, mode):
#    		return os.mkdir(path, mode)
		log.debug("mkdir:" + path + ":")
		if path.startswith("/tags"):
			if path.count('/')==3:   #/tags/personal (or private)/dirname ONLY
				self._mkfileOrDir(path, isDir=True)
				thread.start_new_thread(self.tags_thread, (path,))
			else:
				e = OSError("Not allowed to create directory:%s:"%(path))
				e.errno = EACCES
				raise e
		elif path.startswith("/sets"):
			if path.count('/')==2:  #Only allow creation of new set /sets/newset
				self._mkfileOrDir(path, id=0, isDir=True) 
					#id=0 means that not yet created online
			else:
				e = OSError("Not allowed to create directory:%s:"%(path))
				e.errno = EACCES
				raise e
		else:
			e = OSError("Not allowed to create directory:%s:"%(path))
			e.errno = EACCES
			raise e
			

    	#@-node:mkdir
    	#@+node:utime
    	def utime(self, path, times):
		inode = self.getInode(path)
		inode.atime = times[0]
		inode.mtime = times[1]
#    		return os.utime(path, times)
		return 0
    	#@-node:utime
    	#@+node:open
    	def open(self, path, flags):
		log.info("open: " + path)
		if path.startswith('/sets/'):
			templist = path.split('/')
			ind = templist.index('sets')
			setName = templist[ind+1].split(':')[0]
			templist[2] = setName
			path = '/'.join(templist)
			log.debug("open:After modifying:%s:" % (path))
		
		inode = self.getInode(path)
		if inode.photoId=="NEW": #Just skip if new (i.e. uploading)
			return 0
		if inode.buf=="":	
			log.debug("Retrieving image from flickr: " + inode.photoId)
			inode.buf = str(TransFlickr().getPhoto(inode.photoId))
			inode.size = long(inode.buf.__len__())
			log.debug("Size of image: " + str(inode.size))
		return 0
	    
	#@-node:open
	#@+node:read
	def read(self, path, len, offset):
		log.debug("read:"+ path + ":offset:" + str(offset))
		inode = self.getInode(path)
		if inode.buf == "":
			log.debug("Retrieving image from flickr: " + inode.photoId)
			inode.buf = str(TransFlickr().getPhoto(inode.photoId))
			inode.size = long(inode.buf.__len__())
		sIndex = int(offset)
		ilen = int(len)
    		temp = inode.buf[sIndex:sIndex+ilen]
		if temp.__len__() < len:
			del inode.buf
			inode.buf=""
		return temp
    
	#@-node:read
    	#@+node:write
    	def write(self, path, buf, off):
		log.debug("write:" + path)
		
		# I think its better that you wait when you upload photos. 
		# So, dun start a thread. Do it inline!
#			thread.start_new_thread(TransFlickr().uploadfile, (path, taglist, inode.buf))
		if path.startswith('/tags/'):
			inode = self.getInode(path)
			inode.buf += buf
			if len(buf) < 4096:
				eind = path.rindex('/')
				sind = path.rindex('/',0,eind)
				tags = string.split(path[sind+1:eind], ':')
				tags = [ '"%s"'%(a,) for a in tags]
				tags.append('flickrfs')
				taglist = ' '.join(tags)
				id = TransFlickr().uploadfile(path, taglist, inode.buf, inode.mode)
				del inode.buf
				inode.buf = ""
				inode.photoId = id

		elif path.startswith('/sets/'):
			templist = path.split('/')
			ind = templist.index('sets')
			setnTags = templist[ind+1].split(':')
			setName = setnTags.pop(0)
			templist[2] = setName
			path = '/'.join(templist)
			inode = self.getInode(path)
			inode.buf += buf

			if len(buf) < 4096:
				templist.pop(-1)
				parentPath = '/'.join(templist)
				pinode = self.getInode(parentPath)
				setnTags = [ '"%s"'%(a,) for a in setnTags]
				setnTags.append('flickrfs')
				taglist = ' '.join(setnTags)
				log.debug("Uploading with tags: %s" % (taglist))
				id = TransFlickr().uploadfile(path, taglist, inode.buf, inode.mode)
				del inode.buf
				inode.buf = ""
				inode.photoId = id
				
				#Create set if it doesn't exist online (i.e. if id=0)
				if pinode.setId==0:
					try:
						pinode.setId = TransFlickr().createSet(parentPath, id)
					except:
						e = OSError("Can't create a new set")
						e.errno = EIO
						raise e
				else:
					TransFlickr().put2Set(pinode.setId, id)
				
    		return len(buf)
	#@-node:write

	#@+node:getInode
	def getInode(self, path):
		if self.inodeCache.has_key(path):
			log.debug("Got cached inode: " + path)
			return self.inodeCache[path]
		else:
			log.debug("No inode??? I DIE!!!")
			return None

	#@-node:getInode

	#@+node:release
	def release(self, path, flags):
	        log.debug("flickrfs.py:Flickrfs:release: %s %s" % (path,flags))
	        return 0
	#@-node:release
	
	#@+node:statfs
    	def statfs(self):
        	"""
        Should return a tuple with the following elements in respective order:
	
    	F_BSIZE - Preferred file system block size. (int)
	F_FRSIZE - Fundamental file system block size. (int)
	F_BLOCKS - Total number of blocks in the filesystem. (long)
	F_BFREE - Total number of free blocks. (long)
	F_BAVAIL - Free blocks available to non-super user. (long)
	F_FILES - Total number of file nodes. (long)
	F_FFREE - Total number of free file nodes. (long)
	F_FAVAIL - Free nodes available to non-super user. (long)
	F_FLAG - Flags. System dependent: see statvfs() man page. (int)
	F_NAMEMAX - Maximum file name length. (int)
        Feel free to set any of the above values to 0, which tells
        the kernel that the info is not available.
        """
		log.debug("statfs called")
        	block_size = 1024
		fun_block_size = 1024
        	total_blocks = 0L
        	blocks_free = 0L
		blocks_free_user = 0L
        	files = 0L
        	files_free = 0L
		files_free_user = 0L
		flag = 0
        	namelen = 255
		(max, used) = TransFlickr().getBandwidthInfo()
		if max!=None:
			total_blocks = long(int(max)/block_size)
			blocks_free = long( ( int(max)-int(used) )/block_size)
			blocks_free_user = blocks_free
			log.debug('total blocks:%s'%(total_blocks))
			log.debug('blocks_free:%s'%(blocks_free))
        	return (block_size, fun_block_size, total_blocks, blocks_free, blocks_free_user, \
			files, files_free, files_free_user, namelen)
	#@-node:statfs
	#@+node:fsync
	def fsync(self, path, isfsyncfile):
	        log.debug("flickrfs.py:Flickrfs:fsync: path=" + path + ", isfsyncfile=" + isfsyncfile)
	        return 0
    
	#@-node:fsync

#@+node:mainline
if __name__ == '__main__':
	try:
		server = Flickrfs()
		server.multithreaded = 1;
		server.main()
	except KeyError:
		log.error('Got key error. Exiting...')
		sys.exit(0)
#@-node:mainline
#@-leo
